import depthai as dai

PRIMARY_COLOR = dai.Color(
    float(87 / 255), float(36 / 255), float(232 / 255), float(1.0)
)
TRANSPARENT_PRIMARY_COLOR = dai.Color(
    float(87 / 255), float(36 / 255), float(232 / 255), float(0.2)
)
SECONDARY_COLOR = dai.Color(
    float(240 / 255), float(240 / 255), float(240 / 255), float(1.0)
)
